@@ -6,33 +6,37 @@ import uuid
 
 from .helpers import get_logger, chown_nobody_permissions
 from .core import Connector
-from .connection import Structures
+from .connection import Structures, MessageFields, Misc
 
 DEFAULT_LOGGER_NAME = 'aioconnectors'
 LOGFILE_DEFAULT_NAME = 'aioconnectors.log'
 DEFAULT_LOGGER_LOG_LEVEL = 'INFO'
+DEFAULT_LOGGER_ROTATE = True
 
 class ConnectorManager:
     def __init__(self, config_file_path=None, logger=None, use_default_logger=True, 
-                 default_logger_log_level=DEFAULT_LOGGER_LOG_LEVEL, default_logger_dirpath=Connector.CONNECTOR_FILES_DIRPATH,
+                 default_logger_log_level=DEFAULT_LOGGER_LOG_LEVEL, default_logger_rotate=True,
+                 default_logger_dirpath=Connector.CONNECTOR_FILES_DIRPATH,
                  is_server=True, server_sockaddr=None, use_ssl=Connector.USE_SSL, ssl_allow_all=False, 
                  client_bind_ip=None, certificates_directory_path=None, client_name=None, 
-                 send_message_types=None, recv_message_types=None, connector_files_dirpath=Connector.CONNECTOR_FILES_DIRPATH,
+                 send_message_types=None, recv_message_types=None, subscribe_message_types=None,
+                 connector_files_dirpath=Connector.CONNECTOR_FILES_DIRPATH,
                  disk_persistence_send=Connector.DISK_PERSISTENCE_SEND, 
                  disk_persistence_recv=Connector.DISK_PERSISTENCE_RECV, 
                  max_size_persistence_path=Connector.MAX_SIZE_PERSISTENCE_PATH,
                  file_recv_config=None, debug_msg_counts=Connector.DEBUG_MSG_COUNTS, silent=Connector.SILENT, #use_ack=Connector.USE_ACK,
                  uds_path_receive_preserve_socket=Connector.UDS_PATH_RECEIVE_PRESERVE_SOCKET,
                  uds_path_send_preserve_socket=Connector.UDS_PATH_SEND_PRESERVE_SOCKET,
-                 hook_server_auth_client=None, enable_client_try_reconnect=True,
+                 hook_server_auth_client=None, hook_target_directory=None, enable_client_try_reconnect=True,
                  reuse_server_sockaddr=False, reuse_uds_path_send_to_connector=False, reuse_uds_path_commander_server=False,
                  max_size_file_upload=Connector.MAX_SIZE_FILE_UPLOAD,
                  everybody_can_send_messages=Connector.EVERYBODY_CAN_SEND_MESSAGES,
-                 send_message_types_priorities=None):
+                 send_message_types_priorities=None, pubsub_central_broker=False, proxy=None):
         
         self.connector_files_dirpath = connector_files_dirpath
         self.default_logger_dirpath = default_logger_dirpath
         self.default_logger_log_level = default_logger_log_level
+        self.default_logger_rotate = default_logger_rotate
         if not os.path.isdir(self.connector_files_dirpath):
             os.makedirs(self.connector_files_dirpath)
         if not os.path.isdir(self.default_logger_dirpath):
@@ -41,7 +45,8 @@ class ConnectorManager:
         if not logger:
             if use_default_logger:
                 self.logger = get_logger(logfile_path=os.path.join(self.default_logger_dirpath, LOGFILE_DEFAULT_NAME), 
-                                         logger_name=DEFAULT_LOGGER_NAME, silent=True, level=default_logger_log_level)
+                                         logger_name=DEFAULT_LOGGER_NAME, silent=True, level=default_logger_log_level,
+                                         rotate=default_logger_rotate)
             else:
                 self.logger = get_logger(logfile_path=None)  #dummy logger
         else:
@@ -51,15 +56,20 @@ class ConnectorManager:
                             is_server, server_sockaddr, use_ssl, ssl_allow_all, certificates_directory_path
         self.client_name, self.client_bind_ip = client_name, client_bind_ip
         self.send_message_types, self.recv_message_types = send_message_types, recv_message_types
+        self.pubsub_central_broker = pubsub_central_broker        
+        self.subscribe_message_types = subscribe_message_types
         self.max_size_file_upload = max_size_file_upload
         self.everybody_can_send_messages = everybody_can_send_messages
         self.send_message_types_priorities = send_message_types_priorities
+        self.proxy = proxy
+
         self.disk_persistence_send, self.disk_persistence_recv, self.max_size_persistence_path = \
                             disk_persistence_send, disk_persistence_recv, max_size_persistence_path
         self.file_recv_config, self.debug_msg_counts, self.silent = file_recv_config, debug_msg_counts, silent
         self.uds_path_receive_preserve_socket, self.uds_path_send_preserve_socket = \
                             uds_path_receive_preserve_socket, uds_path_send_preserve_socket
-        self.hook_server_auth_client, self.enable_client_try_reconnect = hook_server_auth_client, enable_client_try_reconnect
+        self.hook_server_auth_client, self.hook_target_directory = hook_server_auth_client, hook_target_directory
+        self.enable_client_try_reconnect = enable_client_try_reconnect
         self.reuse_server_sockaddr, self.reuse_uds_path_send_to_connector, self.reuse_uds_path_commander_server = \
                             reuse_server_sockaddr, reuse_uds_path_send_to_connector, reuse_uds_path_commander_server
 
@@ -78,7 +88,8 @@ class ConnectorManager:
                     if not logger and use_default_logger:
                         self.logger = get_logger(logfile_path=os.path.join(self.default_logger_dirpath,
                                                  LOGFILE_DEFAULT_NAME), logger_name=DEFAULT_LOGGER_NAME, 
-                                                 silent=self.silent, level=self.default_logger_log_level)                            
+                                                 silent=self.silent, level=self.default_logger_log_level,
+                                                 rotate=self.default_logger_rotate)                            
                 except Exception:
                     self.logger.exception('ConnectorManager init config_file_path')
             else:
@@ -97,11 +108,20 @@ class ConnectorManager:
                 self.logger.warning(f'No client_name provided, using {self.client_name} instead')                
             self.source_id = self.client_name
         
+        if self.pubsub_central_broker:
+            if self.recv_message_types is None:
+                self.recv_message_types = []
+            self.recv_message_types.append('_pubsub')        
+        if self.send_message_types is None:
+            self.send_message_types = []
+        self.send_message_types.append('_pubsub')        
+        
         self.connector = Connector(self.logger, is_server=self.is_server, server_sockaddr=self.server_sockaddr, 
                                    use_ssl=self.use_ssl, ssl_allow_all=self.ssl_allow_all,
                                    certificates_directory_path=self.certificates_directory_path, 
                                    client_name=self.client_name, client_bind_ip=self.client_bind_ip,
                                    send_message_types=self.send_message_types, recv_message_types=self.recv_message_types,
+                                   subscribe_message_types=self.subscribe_message_types,
                                    disk_persistence_send=self.disk_persistence_send, 
                                    disk_persistence_recv=self.disk_persistence_recv,
                                    max_size_persistence_path=self.max_size_persistence_path, 
@@ -110,14 +130,16 @@ class ConnectorManager:
                                    connector_files_dirpath=self.connector_files_dirpath, #use_ack=use_ack,
                                    uds_path_receive_preserve_socket=self.uds_path_receive_preserve_socket, 
                                    uds_path_send_preserve_socket=self.uds_path_send_preserve_socket,
-                                   hook_server_auth_client=self.hook_server_auth_client, 
+                                   hook_server_auth_client=self.hook_server_auth_client,
+                                   hook_target_directory=self.hook_target_directory,
                                    enable_client_try_reconnect=self.enable_client_try_reconnect,
                                    reuse_server_sockaddr=self.reuse_server_sockaddr,
                                    reuse_uds_path_send_to_connector=self.reuse_uds_path_send_to_connector,
                                    reuse_uds_path_commander_server=self.reuse_uds_path_commander_server,
                                    max_size_file_upload=self.max_size_file_upload,
                                    everybody_can_send_messages=self.everybody_can_send_messages,
-                                   send_message_types_priorities=self.send_message_types_priorities)        
+                                   send_message_types_priorities=self.send_message_types_priorities,
+                                   pubsub_central_broker=self.pubsub_central_broker, proxy=self.proxy)        
         
             
     async def start_connector(self, delay=None, connector_socket_only=False):        
@@ -126,6 +148,57 @@ class ConnectorManager:
             await asyncio.sleep(delay)                
         self.logger.info('start_connector : '+str(self.source_id))        
         await self.connector.start(connector_socket_only=connector_socket_only)
+        
+        if self.connector.pubsub_central_broker:    #only possible for server
+            self.central_broker_api = ConnectorAPI(config_file_path=self.config_file_path,
+                logger=self.logger.getChild('central_broker_api'),
+                connector_files_dirpath=self.connector_files_dirpath, 
+                is_server=True, server_sockaddr=self.server_sockaddr,
+                uds_path_receive_preserve_socket=self.uds_path_receive_preserve_socket,
+                uds_path_send_preserve_socket=self.uds_path_send_preserve_socket,
+                send_message_types=self.send_message_types, recv_message_types=self.recv_message_types)
+            self.central_broker_api.clients_subscriptions = {}   #key=message_type, value= list of subscribing client ids
+            
+            async def message_received_cb(logger, transport_json , data, binary):
+                try:                
+                    
+                    if MessageFields.MESSAGE_TYPE_PUBLISH not in transport_json:
+                        logger.info(f'central_broker_api message_received_cb {transport_json} and {data}')                        
+                        #extract client data about subscribe_message_types
+                        logger.debug(f'clients_subscriptions before {self.central_broker_api.clients_subscriptions}')
+                        data_json = json.loads(data.decode())
+                        client_id = transport_json[MessageFields.SOURCE_ID]
+                        subscribe_message_types = data_json.get('subscribe_message_types', [])
+                        for message_type in subscribe_message_types:
+                            if message_type not in self.central_broker_api.clients_subscriptions:
+                                self.central_broker_api.clients_subscriptions[message_type] = []
+                            logger.info(f'Subscribing client {client_id} to message type {message_type}')
+                            self.central_broker_api.clients_subscriptions[message_type].append(client_id)
+                        unsubscribe_message_types = data_json.get('unsubscribe_message_types', [])
+                        for message_type in unsubscribe_message_types:
+                            if message_type in self.central_broker_api.clients_subscriptions:
+                                logger.info(f'Unsubscribing client {client_id} to message type {message_type}')
+                                if client_id in self.central_broker_api.clients_subscriptions[message_type]:
+                                    self.central_broker_api.clients_subscriptions[message_type].remove(client_id)
+                        logger.debug(f'clients_subscriptions after {self.central_broker_api.clients_subscriptions}')
+                            
+                    else:
+                        #publish client message to subscribers
+                        client_source = transport_json[MessageFields.SOURCE_ID]
+                        transport_json.pop(MessageFields.SOURCE_ID, None) #use central broker source id, otherwise may break
+                        transport_json.pop(MessageFields.WITH_BINARY, None)                        
+                        message_type_publish = transport_json.pop(MessageFields.MESSAGE_TYPE_PUBLISH)
+                        transport_json[MessageFields.MESSAGE_TYPE] = message_type_publish
+                        for client_destination in self.central_broker_api.clients_subscriptions.get(message_type_publish, []):
+                            logger.debug(f'Publish from {client_source} to {client_destination}')
+                            transport_json[MessageFields.DESTINATION_ID] = client_destination
+                            await self.central_broker_api.send_message(data=data, binary=binary, data_is_json=False,
+                                                                       **transport_json)                        
+                            
+                except Exception:
+                    logger.exception('message_received_cb')
+                    
+            await self.central_broker_api.start_waiting_for_messages(message_type='_pubsub', message_received_cb=message_received_cb)           
 
     async def stop_connector(self, delay=None, connector_socket_only=False, hard=False, shutdown=False, 
                              enable_delete_files=True):
@@ -133,6 +206,9 @@ class ConnectorManager:
             self.logger.info(f'Waiting {delay} seconds before stopping connector : {self.source_id}')
             await asyncio.sleep(delay)        
         self.logger.info('stop_connector : '+str(self.source_id))
+        if self.connector.pubsub_central_broker:    #only possible for server        
+            self.central_broker_api.stop_waiting_for_messages(message_type='_pubsub')           
+        
         await self.connector.stop(connector_socket_only=connector_socket_only, hard=hard, shutdown=shutdown,
                                   enable_delete_files=enable_delete_files)        
         
@@ -164,6 +240,14 @@ class ConnectorManager:
         else:
             res = await self.connector.delete_client_certificate_on_client()
         return res     
+
+    def show_subscribe_message_types(self):
+        self.logger.info(f'{self.source_id} show_subscribe_message_types')
+        return self.connector.show_subscribe_message_types()
+
+    async def set_subscribe_message_types(self, *message_types):
+        self.logger.info(f'{self.source_id} set_subscribe_message_types {message_types}')
+        return self.connector.set_subscribe_message_types(*message_types)
         
     def show_connected_peers(self, dump_result=False):
         self.logger.info(f'{self.source_id} show_connected_peers') 
@@ -190,12 +274,13 @@ class ConnectorManager:
         return self.connector.set_log_level(level)       
 
 class ConnectorBaseTool:
-    def __init__(self, config_file_path=None, logger=None, use_default_logger=True, default_logger_log_level=DEFAULT_LOGGER_LOG_LEVEL,
+    def __init__(self, config_file_path=None, logger=None, use_default_logger=True, default_logger_rotate=DEFAULT_LOGGER_ROTATE,
+                 default_logger_log_level=DEFAULT_LOGGER_LOG_LEVEL,
                  default_logger_dirpath=Connector.CONNECTOR_FILES_DIRPATH, connector_files_dirpath=Connector.CONNECTOR_FILES_DIRPATH, 
                  is_server=False, server_sockaddr=None, client_name=None, 
                  uds_path_receive_preserve_socket=Connector.UDS_PATH_RECEIVE_PRESERVE_SOCKET,
                  uds_path_send_preserve_socket=Connector.UDS_PATH_SEND_PRESERVE_SOCKET,
-                 send_message_types=None, recv_message_types=None):
+                 send_message_types=None, recv_message_types=None, pubsub_central_broker=False):
 
         self.connector_files_dirpath = connector_files_dirpath
         self.default_logger_dirpath = default_logger_dirpath
@@ -205,13 +290,15 @@ class ConnectorBaseTool:
         if not logger:
             if use_default_logger:
                 self.logger = get_logger(logfile_path=os.path.join(self.default_logger_dirpath, LOGFILE_DEFAULT_NAME), 
-                                         logger_name=DEFAULT_LOGGER_NAME, silent=True, level=default_logger_log_level)                
+                                         logger_name=DEFAULT_LOGGER_NAME, silent=True, level=default_logger_log_level,
+                                         rotate=default_logger_rotate)                
             else:
                 self.logger = get_logger(logfile_path=None)  #dummy logger
         else:
             self.logger = logger
         self.is_server, self.server_sockaddr, self.client_name = is_server, server_sockaddr, client_name
         self.send_message_types, self.recv_message_types = send_message_types, recv_message_types
+        self.pubsub_central_broker = pubsub_central_broker
         self.uds_path_send_preserve_socket = uds_path_send_preserve_socket
         self.uds_path_receive_preserve_socket = uds_path_receive_preserve_socket
 
@@ -229,7 +316,8 @@ class ConnectorBaseTool:
                     if not logger and use_default_logger:
                         self.logger = get_logger(logfile_path=os.path.join(self.default_logger_dirpath,
                                                  LOGFILE_DEFAULT_NAME), logger_name=DEFAULT_LOGGER_NAME, 
-                                                 silent=self.silent, level=self.default_logger_log_level)                                                        
+                                                 silent=self.silent, level=self.default_logger_log_level,
+                                                 rotate=self.default_logger_rotate)                                                        
                 except Exception:
                     self.logger.exception('type(self).__name__ init config_file_path')
             else:
@@ -251,10 +339,19 @@ class ConnectorBaseTool:
             self.source_id = self.client_name
         self.reader_writer_uds_path_send = None
         self.message_waiters = {}
+        
+        if self.pubsub_central_broker:
+            if self.recv_message_types is None:
+                self.recv_message_types = []
+            self.recv_message_types.append('_pubsub')        
+        if self.send_message_types is None:
+            self.send_message_types = []
+        #self.send_message_types.append('_pubsub')             
+        
         self.connector = Connector(self.logger, tool_only=True, is_server=self.is_server, 
                             server_sockaddr=self.server_sockaddr, connector_files_dirpath=self.connector_files_dirpath, 
                             client_name=self.client_name, send_message_types=self.send_message_types, 
-                            recv_message_types=self.recv_message_types)
+                            recv_message_types=self.recv_message_types, pubsub_central_broker=self.pubsub_central_broker)
         
 class ConnectorAPI(ConnectorBaseTool):
     '''
@@ -278,31 +375,100 @@ class ConnectorAPI(ConnectorBaseTool):
         self.receive_from_any_connector_owner = receive_from_any_connector_owner
         
     async def send_message_await_response(self, message_type=None, destination_id=None, request_id=None, response_id=None,
-                           data=None, data_is_json=True, binary=None, await_response=False, with_file=None, wait_for_ack=False):
+                           data=None, data_is_json=True, binary=None, await_response=False, with_file=None,
+                           wait_for_ack=False, message_type_publish=None, await_response_timeout=None):
         res = await self.send_message(await_response=True, message_type=message_type, destination_id=destination_id, 
                                       request_id=request_id, response_id=response_id, data=data, 
                                       data_is_json=data_is_json, binary=binary, with_file=with_file, 
-                                      wait_for_ack=wait_for_ack)
+                                      wait_for_ack=wait_for_ack, message_type_publish=message_type_publish,
+                                      await_response_timeout=await_response_timeout)
         return res
 
     def send_message_sync(self, message_type=None, destination_id=None, request_id=None, response_id=None,
-                           data=None, data_is_json=True, binary=None, await_response=False, with_file=None, wait_for_ack=False):
+                           data=None, data_is_json=True, binary=None, await_response=False, with_file=None, 
+                           wait_for_ack=False, message_type_publish=None, await_response_timeout=None, loop=None):
         self.logger.debug(f'send_message_sync of type {message_type}, destination_id {destination_id}, '
                           f'request_id {request_id}, response_id {response_id}')
         
-        loop = asyncio.get_event_loop()
+        loop = loop or asyncio.get_event_loop()
         send_task = self.send_message(message_type=message_type, destination_id=destination_id, 
                                       request_id=request_id, response_id=response_id, data=data, 
                                       data_is_json=data_is_json, binary=binary, await_response=await_response, 
-                                      with_file=with_file, wait_for_ack=wait_for_ack)
+                                      with_file=with_file, wait_for_ack=wait_for_ack,
+                                      message_type_publish=message_type_publish, await_response_timeout=await_response_timeout)
         if loop.is_running():
-            loop.create_task(send_task)
+            return loop.create_task(send_task)
         else:
-            loop.run_until_complete(send_task)    
+            return loop.run_until_complete(send_task)
     
     async def send_message(self, message_type=None, destination_id=None, request_id=None, response_id=None,
                            data=None, data_is_json=True, binary=None, await_response=False, with_file=None, 
-                           wait_for_ack=False):
+                           wait_for_ack=False, message_type_publish=None, await_response_timeout=None, check_chunk_file=True):
+
+        if with_file:
+            src_path = with_file.get('src_path')
+            if src_path and check_chunk_file:
+                file_size = os.path.getsize(src_path)
+                number_of_chunks, last_bytes_size = divmod(file_size, Connector.MAX_SIZE_CHUNK_UPLOAD)
+                if number_of_chunks:
+                    #divide file in chunks of max size MAX_SIZE_CHUNK_UPLOAD, and send each chunk one after the other                    
+                    dst_name = with_file.get('dst_name')
+                    chunk_basepath = self.connector_files_dirpath #os.path.dirname(src_path)
+                    chunk_basename = f'{dst_name}{Misc.CHUNK_INDICATOR}' #f'{dst_name}__aioconnectors_chunk'
+                    try:
+                        override_src_file_sizes = number_of_chunks * [Connector.MAX_SIZE_CHUNK_UPLOAD]
+                        if last_bytes_size:
+                            override_src_file_sizes += [last_bytes_size]
+                        len_override_src_file_sizes = len(override_src_file_sizes)
+                        chunk_names = []
+                        fd = open(src_path, 'rb')
+                        for index, chunk_size in enumerate(override_src_file_sizes):
+                            chunk_name = f'{chunk_basename}_{index+1}_{len_override_src_file_sizes}'
+                            with open(os.path.join(chunk_basepath, chunk_name), 'wb') as fw:
+                                number_of_read_chunks, last_size = divmod(chunk_size, Connector.READ_CHUNK_SIZE)
+                                while number_of_read_chunks:                                    
+                                    number_of_read_chunks -= 1                                    
+                                    chunk_file = fd.read(Connector.READ_CHUNK_SIZE)    
+                                    fw.write(chunk_file)
+                                chunk_file = fd.read(last_size)
+                                fw.write(chunk_file)
+
+                            self.logger.info(f'send_message of type {message_type}, destination_id {destination_id}, '
+                              f'request_id {request_id}, response_id {response_id} creating chunk {chunk_name}')                            
+                            chunk_names.append(chunk_name)
+                            await asyncio.sleep(0)
+                        chunk_file = None
+                    except Exception:
+                        self.logger.exception('send_message chunks')
+                        return False
+                    for index, chunk_name in enumerate(chunk_names):
+                        chunk_name_path = os.path.join(chunk_basepath, chunk_name)                        
+                        with_file['src_path'] = chunk_name_path
+                        with_file['dst_name'] = chunk_name
+                        with_file['chunked'] = [chunk_basename, index+1, len_override_src_file_sizes]
+                        res = await self.send_message(message_type=message_type, destination_id=destination_id,
+                                                      request_id=request_id, response_id=response_id,
+                                                      data=data, data_is_json=data_is_json, binary=binary,
+                                                      await_response=await_response, with_file=with_file, 
+                                                      wait_for_ack=wait_for_ack, message_type_publish=message_type_publish,
+                                                      await_response_timeout=await_response_timeout, check_chunk_file=False)
+                        if os.path.exists(chunk_name_path):
+                            if await_response or wait_for_ack:
+                                try:
+                                    self.logger.info(f'send_message of type {message_type}, destination_id {destination_id}, '
+                                  f'request_id {request_id}, response_id {response_id} deleting chunk {chunk_name_path}')
+                                    
+                                    os.remove(chunk_name_path)
+                                except Exception:
+                                    self.logger.exception(f'send_message of type {message_type}, destination_id {destination_id}, '
+                                  f'request_id {request_id}, response_id {response_id} deleting chunk {chunk_name_path}')
+                            else:
+                                self.logger.warning(f'send_message of type {message_type}, destination_id {destination_id}, '
+                                  f'request_id {request_id}, response_id {response_id} leaving undeleted chunk {chunk_name_path}')                                
+                                
+                        if not res:
+                            return res
+                    return res                       
 
         try:  
             
@@ -316,7 +482,7 @@ class ConnectorAPI(ConnectorBaseTool):
             message_bytes = self.connector.pack_message(data=data, message_type=message_type, source_id=self.source_id,
                                    destination_id=destination_id, request_id=request_id, response_id=response_id, 
                                    binary=binary, await_response=await_response, with_file=with_file, 
-                                   wait_for_ack=wait_for_ack)
+                                   wait_for_ack=wait_for_ack, message_type_publish=message_type_publish)
 
             send_message_lock_internally_acquired = False
             if self.uds_path_send_preserve_socket and not await_response:
@@ -399,8 +565,16 @@ class ConnectorAPI(ConnectorBaseTool):
             except Exception:
                 self.logger.exception('send_message writer drain')
             #beware to not lock the await_response recv_message with send_message_lock
-            if await_response:            
-                the_response = await self.recv_message(reader, writer)
+            if await_response:
+                if await_response_timeout is not None:
+                    try:
+                        the_response = await asyncio.wait_for(self.recv_message(reader, writer), timeout=await_response_timeout)
+                    except asyncio.TimeoutError:
+                        self.logger.warning(f'send_message : await_response_timeout error ({await_response_timeout} s)')
+                        writer.close()
+                        return False                      
+                else:
+                    the_response = await self.recv_message(reader, writer)
             self.logger.debug('send_message finished sending')                    
             if await_response:
                 writer.close()
@@ -419,6 +593,22 @@ class ConnectorAPI(ConnectorBaseTool):
             self.logger.exception('send_data')
             return False
 
+    async def publish_message(self, message_type=None, destination_id=None, request_id=None, response_id=None,
+                           data=None, data_is_json=True, binary=None, await_response=False, with_file=None, 
+                           wait_for_ack=False):
+        res = await self.send_message(message_type='_pubsub', message_type_publish=message_type, destination_id=destination_id, 
+                                      request_id=request_id, response_id=response_id, data=data, 
+                                      data_is_json=data_is_json, binary=binary, await_response=await_response,
+                                      with_file=with_file, wait_for_ack=wait_for_ack)
+        return res
+    
+    def publish_message_sync(self, message_type=None, destination_id=None, request_id=None, response_id=None,
+                           data=None, data_is_json=True, binary=None, await_response=False, with_file=None, wait_for_ack=False):
+        res = self.send_message_sync(message_type='_pubsub', message_type_publish=message_type,
+                                     destination_id=destination_id, request_id=request_id, response_id=response_id,
+                                     data=data, data_is_json=data_is_json, binary=binary, await_response=await_response,
+                                     with_file=with_file, wait_for_ack=wait_for_ack)
+        return res
             
     async def recv_message(self, reader, writer):
         try:
@@ -566,14 +756,14 @@ class ConnectorRemoteTool(ConnectorBaseTool):
         response = await self.send_command(cmd='restart', kwargs={'hard':hard, 'sleep_between':sleep_between})
         return response
                 
-    async def delete_client_certificate(self, client_id=None, remove_only_symlink=False):
+    async def delete_client_certificate(self, client_id=None, remove_only_symlink=False, restart_client=True):
         self.logger.info(f'{self.source_id} delete_client_certificate {client_id}')                 
         if self.is_server:
             response = await self.send_command(cmd='delete_client_certificate_on_server', 
                                                kwargs={'client_id':client_id, 'remove_only_symlink':remove_only_symlink})
             return response
         else:
-            response = await self.send_command(cmd='delete_client_certificate_on_client', kwargs={})
+            response = await self.send_command(cmd='delete_client_certificate_on_client', kwargs={'restart_client':restart_client})
             return response
         
     async def disconnect_client(self, client_id=None):
@@ -589,6 +779,16 @@ class ConnectorRemoteTool(ConnectorBaseTool):
         response = await self.send_command(cmd='delete_previous_persistence_remains__sync', kwargs={})        
         return response
     
+    async def show_subscribe_message_types(self):
+        self.logger.info(f'{self.source_id} show_subscribe_message_types')         
+        response = await self.send_command(cmd='show_subscribe_message_types__sync', kwargs={})        
+        return response
+    
+    async def set_subscribe_message_types(self, *message_types):
+        self.logger.info(f'{self.source_id} set_subscribe_message_types {message_types}')         
+        response = await self.send_command(cmd='set_subscribe_message_types', kwargs={'message_types':message_types})
+        return response
+    
     async def show_connected_peers(self):
         self.logger.info(f'{self.source_id} show_connected_peers')         
         response = await self.send_command(cmd='show_connected_peers__sync', kwargs={})        
