@@ -7,6 +7,7 @@ import zipfile
 import subprocess
 import sys
 import json
+import re
 
 import aioconnectors
 
@@ -61,7 +62,8 @@ def cli(logger=None):
             if os.path.exists(active_connectors_path):
                 with open(active_connectors_path, 'r') as fd:
                     set_active_connectors = json.load(fd)    
-                dict_connector_names = {str(index):connector_name for index,connector_name in enumerate(set_active_connectors)}
+                dict_connector_names = {str(index):connector_name for index,connector_name in \
+                                            enumerate(sorted(set_active_connectors))}
                 display_dict(dict_connector_names)        
                 print('\nPlease type the connector number you would like to run, or')
         except Exception as exc:
@@ -97,7 +99,7 @@ def cli(logger=None):
         clearscreen()
         list_cmds = ['start', 'stop gracefully', 'stop hard', 'restart', 'show_connected_peers', 
                      'ignore_peer_traffic', 'peek_queues', 'delete_client_certificate', 'disconnect_client',
-                     'show_log_level', 'set_log_level']
+                     'show_log_level', 'set_log_level', 'show_subscribe_message_types', 'set_subscribe_message_types']
         dict_cmds = {str(index):cmd for index,cmd in enumerate(list_cmds)}
         display_dict(dict_cmds, connector=server_sockaddr or client_name)        
         res = input('\nPlease type the command number you would like to run, or q to quit\n')
@@ -215,6 +217,32 @@ def cli(logger=None):
                     task = loop.create_task(connector_remote_tool.set_log_level(new_level))
                     loop.run_until_complete(task)
                     print(task.result().decode())       
+                elif the_cmd == 'show_subscribe_message_types':
+                    if is_server:
+                        print('Only available for clients')
+                    else:
+                        task = loop.create_task(connector_remote_tool.show_subscribe_message_types())
+                        loop.run_until_complete(task)
+                        print(task.result().decode())     
+                elif the_cmd == 'set_subscribe_message_types':
+                    if is_server:
+                        print('Only available for clients')
+                    else:                    
+                        print('Current subscribed message types are :')
+                        task = loop.create_task(connector_remote_tool.show_subscribe_message_types())
+                        loop.run_until_complete(task)
+                        print(task.result().decode())
+                        res = input('\nPlease type the list of all message types you would like to subscribe, or q to quit\n')
+                        clearscreen()                    
+                        if res == 'q':
+                            break
+                        new_message_types = res.split()                        
+                        res2 = input(f'\nAre you sure you want to subscribe to these message types : {new_message_types} ? y/n\n')
+                        if res2.lower() != 'y':
+                            break      
+                        task = loop.create_task(connector_remote_tool.set_subscribe_message_types(*new_message_types))
+                        loop.run_until_complete(task)
+                        print(task.result().decode())   
                     
             display_dict(dict_cmds, connector=server_sockaddr or client_name)                    
             res = input('\nPlease type the command number you would like to run, or q to quit\n')                    
@@ -282,6 +310,39 @@ def test_send_messages(config_file_path, logger=None):
         task_send.cancel()
         print('test_send_messages stopped !')
         
+def test_publish_messages(config_file_path, logger=None):
+    if not logger:
+        logger = aioconnectors.get_logger(logger_name='test_publish_messages', first_run=True)    
+    logger.info('Creating connector api with config file '+config_file_path)
+    connector_api = aioconnectors.ConnectorAPI(config_file_path=config_file_path)
+    destination_id = None
+    if connector_api.is_server:
+        destination_id = input('\nPlease type the name of your remote client\n')
+        
+    loop = asyncio.get_event_loop()
+    
+    async def send_messages(destination_id):
+        index = 0
+        while True:
+            index += 1
+            for message_type in connector_api.send_message_types:
+                if message_type == '_pubsub':
+                    continue
+                print(f'PUBLISHING MESSAGE to peer {destination_id or connector_api.server_sockaddr} of type '
+                      f'{message_type} and index {index}')
+                response = await connector_api.publish_message(data=f'"TEST_MESSAGE {str(index)*5}"', data_is_json=False,
+                                                                                  destination_id=destination_id,
+                                                        message_type=message_type, await_response=False, request_id=index)
+                                                        #response_id=None, binary=b'\x01\x02\x03\x04\x05', with_file=None, wait_for_ack=False)        
+            await asyncio.sleep(2)
+                    
+    task_send = loop.create_task(send_messages(destination_id))
+    try:
+        loop.run_forever()
+    except:
+        task_send.cancel()
+        print('test_publish_messages stopped !')
+        
 def ping(config_file_path, logger=None):
     #lets a connector ping a remote connector peer
     if not logger:
@@ -325,7 +386,7 @@ def chat(args, logger=None):
         logger = aioconnectors.get_logger(logger_name='chat', first_run=True)
     custom_prompt = 'aioconnectors>> '        
     chat_client_name = 'chat_client'
-    CONNECTOR_FILES_DIRPATH = '/tmp/aioconnectors'
+    CONNECTOR_FILES_DIRPATH = aioconnectors.get_tmp_dir()
     delete_connector_dirpath_later = not os.path.exists(CONNECTOR_FILES_DIRPATH)
     is_server = not args.target
     accept_all_clients = args.accept
@@ -362,21 +423,38 @@ def chat(args, logger=None):
         if '.' not in listening_ip:
             listening_ip = aioconnectors.iface_to_ip(listening_ip)
         server_sockaddr = (listening_ip, int(args.port or 0) or aioconnectors.core.Connector.SERVER_ADDR[1])
+        if listening_ip == '0.0.0.0':
+            listening_addresses = show_up_ips()
+        else:
+            listening_addresses = [listening_ip]
+            
+        print(f'Chat Server listening on addresses {listening_addresses[:5]}, and port {server_sockaddr[1]}\n')
         connector_files_dirpath = CONNECTOR_FILES_DIRPATH
-        aioconnectors.ssl_helper.create_certificates(logger, certificates_directory_path=connector_files_dirpath)            
+        aioconnectors.ssl_helper.create_certificates(logger, certificates_directory_path=connector_files_dirpath)
+        
+        def hook_target_directory_any(transport_json):
+            #this is just for testing the hook feature
+            #this hook is simple and not really needed, we could have used instead :
+            #{'target_directory':os.path.join(cwd,'{source_id')})}
+            source_id = transport_json['source_id']
+            return source_id      
+        
         connector_manager = aioconnectors.ConnectorManager(is_server=True, server_sockaddr=server_sockaddr, 
                                                            use_ssl=True, ssl_allow_all=True,
                                                            connector_files_dirpath=connector_files_dirpath, 
                                                            certificates_directory_path=connector_files_dirpath,
                                                            send_message_types=['any'], recv_message_types=['any'], 
                                                            file_recv_config={'any': {'target_directory':cwd}},
+                                                           #file_recv_config={'any': {'target_directory':os.path.join(cwd,'{source_id}')}},                                                           
                                                            hook_server_auth_client=None if accept_all_clients else \
-                                                           AuthClient.authenticate_client)
+                                                           AuthClient.authenticate_client, #)
+                                                           hook_target_directory={'any':hook_target_directory_any})
                     
         connector_api = aioconnectors.ConnectorAPI(is_server=True, server_sockaddr=server_sockaddr, 
                                                    connector_files_dirpath=connector_files_dirpath,
                                                            send_message_types=['any'], recv_message_types=['any'], 
-                                                           default_logger_log_level='INFO')
+                                                           default_logger_log_level='INFO',
+                                                           default_logger_rotate=True)
         destination_id = chat_client_name
     else:
         server_sockaddr = (args.target, args.port or aioconnectors.core.Connector.SERVER_ADDR[1])
@@ -394,7 +472,8 @@ def chat(args, logger=None):
                                                    connector_files_dirpath=connector_files_dirpath, 
                                                    client_name=chat_client_name,
                                                    send_message_types=['any'], recv_message_types=['any'], 
-                                                   default_logger_log_level='INFO')
+                                                   default_logger_log_level='INFO',
+                                                   default_logger_rotate=True)
         destination_id = None
         
         
@@ -565,3 +644,18 @@ def chat(args, logger=None):
     if delete_connector_dirpath_later and os.path.exists(connector_files_dirpath):
         shutil.rmtree(connector_files_dirpath)
     
+def show_up_ips():
+    #tries to return list of up ipv4 ips (just for display)
+    try:
+        IPADDR_REGEX = re.compile('(?P<ipaddr>[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)')
+        cmd = ['ip', '-4', '-br', '-f', 'inet', 'addr', 'show']
+        ifconfig_lines = subprocess.check_output(cmd, encoding='utf8', timeout=5).splitlines()
+        addresses = []
+        for line in ifconfig_lines:
+            if 'UP' in line:
+                res = IPADDR_REGEX.search(line)
+                if res:
+                    addresses.append(res.group('ipaddr'))
+        return addresses
+    except Exception:
+        return []
\ No newline at end of file
