@@ -33,7 +33,7 @@ from .helpers import full_path, chown_file, CustomException, PYTHON_GREATER_37
 DEBUG_SHOW_DATA = False
 
 class MessageFields:
-    MESSAGE_TYPE = 'message_type'    #'_ssl', '_ack', '_ping', '_handshake_ssl', '_handshake_no_ssl', <user-defined>, ...
+    MESSAGE_TYPE = 'message_type'    #'_ssl', '_ack', '_ping', '_handshake_ssl', '_handshake_no_ssl', '_pubsub', <user-defined>, ...
     SOURCE_ID = 'source_id'    #str
     DESTINATION_ID = 'destination_id'    #str
     REQUEST_ID = 'request_id'    #int
@@ -44,12 +44,15 @@ class MessageFields:
     WITH_FILE = 'with_file'    
     TRANSPORT_ID = 'transport_id'    #int
     WAIT_FOR_ACK = 'wait_for_ack'    #boolean
+    MESSAGE_TYPE_PUBLISH = 'message_type_publish'
     
 class Structures:
     MSG_4_STRUCT = Struct('I')    #4
     MSG_2_STRUCT = Struct('H')    #2
 #MSG_LENGTH_STRUCT = Struct('Q')    #8
     
+class Misc:
+    CHUNK_INDICATOR = '__aioconnectors_chunk'
     
 class FullDuplex:
 
@@ -122,6 +125,15 @@ class FullDuplex:
         
     async def stop(self, client_wait_for_reconnect=False, hard=False):
         try:
+            
+            if not self.connector.is_server:
+                if self.connector.subscribe_message_types:
+                    try:
+                        await self.send_message(message_type='_pubsub',
+                                    data=json.dumps({'unsubscribe_message_types':self.connector.subscribe_message_types}))
+                    except Exception:
+                        self.logger.exception(self.connector.source_id+' stop')                    
+            
             self.is_stopping = True
             self.logger.info(f'{self.connector.source_id} stop FullDuplex to peer {self.peername}')
             self.logger.info(f'queue_recv size {self.connector.queue_recv.qsize()} , queue_send size '
@@ -131,10 +143,15 @@ class FullDuplex:
             #we never stop listening to uds queue_send_to_connector,
             #but we remove peername from queue_send so that it stops writing messages into queue_send
             #this has been done already in stop_nowait_for_persistence in case of persistence, but need to be done here otherwise
+            
+            #stop_nowait_for_persistence is not necessarily called during handle_ssl and no_ssl_handshake, so 
+            #pop full_duplex_connections always : at worst it will try to pop again.
+            self.connector.full_duplex_connections.pop(self.peername, None)                
+            
             if not self.connector.disk_persistence:            
                 #this has been done already in stop_nowait_for_persistence if disk_persistence
                 self.logger.info(f'{self.connector.source_id} stop, now deleting queue_send of peer {self.peername}')
-                self.connector.full_duplex_connections.pop(self.peername, None)                
+#                self.connector.full_duplex_connections.pop(self.peername, None)                
                 queue_send = self.connector.queue_send.pop(self.peername, None)
                 if queue_send:
                     if hard:
@@ -227,7 +244,7 @@ class FullDuplex:
                 peer_identification_finished = True
                 
             if peername+'_incoming' in self.connector.tasks:
-                #problem here in case of server, after client without ssh had its peername replaced. 
+                #problem here in case of server, after client without ssl had its peername replaced. 
                 #we won't detect the redundant connection
                 self.logger.warning('peername : '+str(peername)+' already connected : Disconnecting and reconnecting...')
                 self.connector.cancel_tasks(task_names=[peername+'_incoming', peername+'_outgoing'])
@@ -298,6 +315,14 @@ class FullDuplex:
                     self.logger.exception(self.connector.source_id+' handle_incoming_connection')                    
                     return                   
         
+            if self.connector.subscribe_message_types:
+                try:
+                    await self.send_message(message_type='_pubsub',
+                                data=json.dumps({'subscribe_message_types':self.connector.subscribe_message_types}))
+                except Exception:
+                    self.logger.exception(self.connector.source_id+' handle_incoming_connection')                    
+                    return                   
+        
         while True:
             try:            
                 self.logger.debug(self.connector.source_id+' handle_incoming_connection waiting for message')
@@ -403,9 +428,20 @@ class FullDuplex:
                     #dict {'src_path':, 'dst_name':, 'dst_type':, 'binary_offset':, 'owner':'user:group'}
                     if with_file:
                         file_recv_config = self.connector.file_recv_config.get(with_file.get('dst_type'))
-                        dst_dirpath = file_recv_config['target_directory'].format(**transport_json)
                         file_owner = file_recv_config.get('owner')
                         override_existing = file_recv_config.get('override_existing', False)
+
+                        dst_dirpath = file_recv_config['target_directory'].format(**transport_json)
+                        try:
+                            if self.connector.hook_target_directory:
+                                hook_target_directory = self.connector.hook_target_directory.get(with_file['dst_type'])                                
+                                #user should have manually defined hook_target_directory(transport_json) beforehand
+                                if hook_target_directory:
+                                    dst_dirpath = os.path.join(dst_dirpath, hook_target_directory(transport_json))
+                        except Exception:
+                            self.logger.exception('hook_target_directory')
+                        self.logger.info(f'{self.connector.source_id} handle_incoming_connection from peer '
+                                        f'{self.peername} storing file into {dst_dirpath}')
                         
                         if dst_dirpath:
                             binary_offset = 0
@@ -464,8 +500,57 @@ class FullDuplex:
                                                                         
                                     with open(dst_fullpath, 'wb') as fd:
                                         fd.write(binary[binary_offset:])
-                                    if file_owner:
-                                        chown_file(dst_fullpath, *file_owner, self.logger)                                        
+                                    if self.connector.pubsub_central_broker:
+                                        with_file['src_path'] = dst_fullpath    #get ready to resend file to subscribers
+                                    else:
+                                        if file_owner:
+                                            chown_file(dst_fullpath, *file_owner, self.logger)
+                                        try:
+                                            chunked = with_file.get('chunked')
+                                            if chunked:
+                                                self.logger.debug(f'{self.connector.source_id} handle_incoming_connection from '
+                                                                f'peer {self.peername} received with_file with chunked '
+                                                                f'{chunked}')
+                                                #chunked looks like : [chunk_basename, index+1, len_override_src_file_sizes]
+                                                chunk_basename, chunk_index, number_of_chunks = chunked
+                                                dst_dir_files = [thefile for thefile in os.listdir(dir_to_test) if \
+                                                                 thefile.startswith(chunk_basename)]
+                                                if len(dst_dir_files) == number_of_chunks:
+                                                    files_to_unify = sorted([os.path.join(dir_to_test, thefile) for thefile \
+                                                                       in dst_dir_files])                                                    
+                                                    dst_filename = chunk_basename.split(Misc.CHUNK_INDICATOR)[0]
+                                                    dst_filename_fullpath = full_path(os.path.join(dst_dirpath, dst_filename))
+                                                    unify_chunks = True
+                                                    if os.path.exists(dst_filename_fullpath):
+                                                        if not override_existing:
+                                                            self.logger.warning(f'{self.connector.source_id} handle_incoming_connection from '
+                                                                            f'peer {self.peername} trying to override existing file '
+                                                                            f'{dst_filename_fullpath}, ignoring...')
+                                                            unify_chunks = False
+                                                    if unify_chunks:
+                                                        self.logger.info(f'{self.connector.source_id} handle_incoming_connection from '
+                                                                        f'peer {self.peername} unifying chunks {files_to_unify} '
+                                                                        f'into file {dst_filename_fullpath}')                                                    
+                                                        fdunify = open(dst_filename_fullpath, 'wb')                                            
+                                                        for thefile in files_to_unify:
+                                                            with open(thefile, 'rb') as fdread:
+                                                                ffread = True
+                                                                while ffread:
+                                                                    ffread = fdread.read(self.connector.READ_CHUNK_SIZE)
+                                                                    fdunify.write(ffread)
+                                                            await asyncio.sleep(0)
+                                                        ffread = None
+                                                        fdunify.close()
+                                                    #delete chunks
+                                                    for thefile in files_to_unify:
+                                                        self.logger.info(f'{self.connector.source_id} handle_incoming_connection from '
+                                                                        f'peer {self.peername} deleting chunk {thefile}')
+                                                        os.remove(thefile)
+                                                    
+                                        except Exception:
+                                            self.logger.exception(f'{self.connector.source_id} from peer {self.peername} '
+                                                                  'chunked exception')
+                                            
                                 #remove file from binary, whether having written it to dst_fullpath or not. To prevent bloating
                                 binary = binary[:binary_offset]
                                 if len(binary) == 0:
@@ -703,7 +788,7 @@ class FullDuplex:
 
     #4|2|json|4|data|4|binary
     async def send_message(self, message=None, message_type=None, source_id=None, destination_id=None, request_id=None,
-                           response_id=None, transport_json=None, data=None, binary=None):
+                           response_id=None, transport_json=None, data=None, binary=None, message_type_publish=None):
         try:
 #            async with self.lock_connection:       
             update_msg_counts = True                   
@@ -730,7 +815,7 @@ class FullDuplex:
                 message = self.connector.pack_message(message_type=message_type, source_id=source_id, 
                                                       destination_id=destination_id, request_id=request_id,
                                                       response_id=response_id, transport_json=transport_json,
-                                                      data=data, binary=binary)
+                                                      data=data, binary=binary, message_type_publish=message_type_publish)
                 if DEBUG_SHOW_DATA:
                     self.logger.info('send_message full message ready to send : '+str(message))    
 
