@@ -6,11 +6,13 @@ import weakref
 import os
 import ssl
 import uuid
+import re
 from copy import deepcopy
 from time import time
+from base64 import b64encode
 
 from .connection import FullDuplex, DEBUG_SHOW_DATA, MessageFields, Structures
-from .helpers import full_path, chown_nobody_permissions, iface_to_ip
+from .helpers import full_path, chown_nobody_permissions, iface_to_ip, get_tmp_dir
 from .ssl_helper import SSL_helper
 
 
@@ -19,16 +21,18 @@ class Connector:
     #default values configurable at __init__
     SERVER_ADDR =  ('127.0.0.1',10673)
     USE_SSL = True
-    CONNECTOR_FILES_DIRPATH = '/tmp/aioconnectors'
+    CONNECTOR_FILES_DIRPATH = get_tmp_dir()
     DISK_PERSISTENCE_SEND = False    #can be boolean, or list of message types having disk persistence enabled
     #RAM_PERSISTENCE cannot be true in the current implementation, since queue_send[peername] doesn't exist anymore in disconnected mode
     #however the code still exists partially, in case of future change
     PERSISTENCE_SEND_FILE_NAME = 'connector_disk_persistence_send_from_{}_to_peer'
     DISK_PERSISTENCE_RECV = False
     PERSISTENCE_RECV_FILE_NAME = 'connector_disk_persistence_recv_by_{}'    
-    MAX_SIZE_PERSISTENCE_PATH = 1_000_000_000 #1gb
-    MAX_SIZE_FILE_UPLOAD = 1_000_000_000 #1gb
-    READ_CHUNK_SIZE = 1024
+    MAX_SIZE_FILE_UPLOAD = 8_589_930_194 #8gb
+    MAX_SIZE_CHUNK_UPLOAD = 1_073_741_824 #1gb    
+    MAX_SIZE_PERSISTENCE_PATH = 1_073_741_824 #1gb
+    READ_CHUNK_SIZE = 104_857_600 #100mb    
+    READ_PERSISTENCE_CHUNK_SIZE = 1024
     UDS_PATH_RECEIVE_PRESERVE_SOCKET = True
     UDS_PATH_SEND_PRESERVE_SOCKET = True    
     SILENT=True    
@@ -68,14 +72,15 @@ class Connector:
                  use_ssl=USE_SSL, ssl_allow_all=False, certificates_directory_path=None, 
                  disk_persistence_send=DISK_PERSISTENCE_SEND,
                  disk_persistence_recv=DISK_PERSISTENCE_RECV, max_size_persistence_path=MAX_SIZE_PERSISTENCE_PATH, #use_ack=USE_ACK,
-                 send_message_types=None, recv_message_types=None, tool_only=False, file_recv_config=None,
+                 send_message_types=None, recv_message_types=None, subscribe_message_types=None,
+                 tool_only=False, file_recv_config=None,
                  debug_msg_counts=DEBUG_MSG_COUNTS, silent=SILENT, connector_files_dirpath = CONNECTOR_FILES_DIRPATH,
                  uds_path_receive_preserve_socket=UDS_PATH_RECEIVE_PRESERVE_SOCKET,
                  uds_path_send_preserve_socket=UDS_PATH_SEND_PRESERVE_SOCKET,
-                 hook_server_auth_client=None, enable_client_try_reconnect=True,
+                 hook_server_auth_client=None, hook_target_directory=None, enable_client_try_reconnect=True,
                  reuse_server_sockaddr=False, reuse_uds_path_send_to_connector=False, reuse_uds_path_commander_server=False,
                  max_size_file_upload=MAX_SIZE_FILE_UPLOAD, everybody_can_send_messages=EVERYBODY_CAN_SEND_MESSAGES,
-                 send_message_types_priorities=None):
+                 send_message_types_priorities=None, pubsub_central_broker=False, proxy=None):
         
         self.logger = logger.getChild('server' if is_server else 'client')
         if tool_only:
@@ -108,12 +113,16 @@ class Connector:
             self.reuse_server_sockaddr = reuse_server_sockaddr
             self.reuse_uds_path_send_to_connector = reuse_uds_path_send_to_connector
             self.reuse_uds_path_commander_server = reuse_uds_path_commander_server
-            
+            self.hook_target_directory = hook_target_directory
             if self.is_server:
                 self.source_id = str(self.server_sockaddr)
                 self.logger.info('Server has source id : '+self.source_id)                
                 self.alnum_source_id = '_'.join([self.alnum_name(el) for el in self.source_id.split()])
                 self.hook_server_auth_client = hook_server_auth_client
+                if hook_server_auth_client:
+                    self.logger.info(f'Connector Server {self.source_id} has a hook_server_auth_client')
+                if hook_target_directory:
+                    self.logger.info(f'Connector Server {self.source_id} has a hook_target_directory')
 
                 if send_message_types is None:
                     send_message_types = self.DEFAULT_MESSAGE_TYPES
@@ -138,14 +147,26 @@ class Connector:
                 self.client_bind_ip = client_bind_ip
                 self.source_id = client_name
                 self.logger.info('Client has source id : '+self.source_id)   
+                if hook_target_directory:
+                    self.logger.info(f'Connector Client {self.source_id} has a hook_target_directory')                
                 self.alnum_source_id = self.alnum_name(self.source_id)                            
                 self.enable_client_try_reconnect = enable_client_try_reconnect
+                self.proxy = proxy or {}
+                self.inside_end_sockpair = None
+                if self.proxy.get('enabled'):
+                    #proxy can be like {'enabled':True, 'address':'1.2.3.4 or bla.com', 'port':'22',
+                    #'ssl_server':False, 'authorization':''},
+                    #'authorization' can also be like : {'username':<username>, 'password':<password>}
+                    self.logger.info(f'Client {self.source_id} will use proxy {str(self.proxy)}')
                 
                 if send_message_types is None:
                     send_message_types = self.DEFAULT_MESSAGE_TYPES
                 if recv_message_types is None:
                     recv_message_types = self.DEFAULT_MESSAGE_TYPES
+                if subscribe_message_types is None:
+                    subscribe_message_types = []
                 self.send_message_types, self.recv_message_types = send_message_types, recv_message_types
+                self.subscribe_message_types = subscribe_message_types
                 
                 self.uds_path_send_to_connector = os.path.join(self.connector_files_dirpath,
                                                     self.UDS_PATH_SEND_TO_CONNECTOR_CLIENT.format(self.alnum_source_id))
@@ -193,6 +214,7 @@ class Connector:
                 self.send_message_types_priorities = send_message_types_priorities
                 if not self.send_message_types_priorities:
                     self.send_message_types_priorities = {}
+                self.pubsub_central_broker = pubsub_central_broker if self.is_server else False
                 self.max_size_file_upload = max_size_file_upload
                 self.disk_persistence = disk_persistence_send
                 self.persistence_path = os.path.join(self.connector_files_dirpath,
@@ -264,8 +286,14 @@ class Connector:
                 self.logger.info('msg_counts : '+str(self.msg_counts))
                 self.previous_msg_counts = deepcopy(self.msg_counts)
             queues_stats = self.peek_queues()
-            if queues_stats['queue_recv'] or any(queues_stats['queue_send'].values()):
-                self.logger.info('peek_queues : '+json.dumps(queues_stats, indent=4, sort_keys=True))
+            queues_stats_display = {}            
+            if queues_stats['queue_recv']:
+                queues_stats_display['queue_recv'] = queues_stats['queue_recv']                
+            if any(queues_stats['queue_send'].values()):
+                queues_stats_display['queue_send'] = {key:value for (key,value) in \
+                                            queues_stats['queue_send'].items() if value}
+            if queues_stats_display:
+                self.logger.info('peek_queues : '+json.dumps(queues_stats_display, indent=4, sort_keys=True))
                 
     def alnum_name(self, name):
         return ''.join([str(letter) for letter in name if str(letter).isalnum()])
@@ -319,11 +347,20 @@ class Connector:
                 if '.' not in server_sockaddr_addr:
                     #this is useful only when client and server are on the same machine
                     server_sockaddr_addr = iface_to_ip(server_sockaddr_addr)
-                await asyncio.wait_for(self.loop.sock_connect(self.sock,
-                                            (server_sockaddr_addr, self.server_sockaddr[1])), timeout=2)                                
-                self.logger.info(f'Created socket for {self.source_id} with info {str(self.sock.getsockname())} '
-                                 f'to peer {self.sock.getpeername()}')
-                
+                    
+                if not self.proxy.get('enabled'):
+                    await asyncio.wait_for(self.loop.sock_connect(self.sock,
+                                                (server_sockaddr_addr, self.server_sockaddr[1])), timeout=2)                                
+                    self.logger.info(f'Created socket for {self.source_id} with info {str(self.sock.getsockname())} '
+                                     f'to peer {self.sock.getpeername()}')
+                else:
+                    #only for client
+                    await asyncio.wait_for(self.loop.sock_connect(self.sock,
+                                                (self.proxy['address'], self.proxy['port'])), timeout=2)                                
+                    self.logger.info(f'Created socket for {self.source_id} with info {str(self.sock.getsockname())} '
+                                     f'to proxy {self.sock.getpeername()}')
+                    await self.proxy_connect(self.server_sockaddr, server_sockaddr_addr=server_sockaddr_addr)
+                    
                 self.tasks['run_client'] = self.loop.create_task(self.run_client())    
                 #self.logger.info('ALL TASKS : '+str(self.tasks['run_client'].all_tasks()))            
               
@@ -359,6 +396,75 @@ class Connector:
                 raise
             return
         
+    async def proxy_connect(self, server_sockaddr, server_sockaddr_addr=None):
+        #request : CONNECT bla.com:22 HTTP/1.1
+        #Proxy-Authorization: Basic <b64encode(username:password)>
+        #response : HTTP/1.1 200 OK
+        try:
+            proxy_msg = f"CONNECT {server_sockaddr_addr or server_sockaddr[0]}:{server_sockaddr[1]} HTTP/1.1"
+            authorization = self.proxy.get('authorization', None)
+            if authorization:
+                #not tested
+                proxy_msg_cont = "\r\nProxy-Authorization: basic " + \
+                b64encode((authorization['username']+':'+authorization['password']).encode()).decode()
+                proxy_msg += proxy_msg_cont
+            proxy_msg += "\r\n\r\n"
+            proxy_msg = proxy_msg.encode()
+            self.logger.info(f'Trying to connect through proxy with : {proxy_msg}')
+            
+            if self.proxy.get('ssl_server', None):
+                self.logger.info('Proxy server listens with SSL')
+                context = ssl.create_default_context()
+                #ssl wrap the client/proxy socket
+                self.sock = context.wrap_socket(self.sock)
+                #create a sockpair :
+                #inside end self.inside_end_sockpair will be used by run_client instead of self.sock
+                #outside end should be forwarded to/from self.sock in 2 new tasks appended to run_client task
+                self.inside_end_sockpair, self.outside_end_sockpair = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM)
+                if not isinstance(self.tasks['run_client'], list):
+                    self.tasks['run_client'] = [self.tasks['run_client']]
+                self.tasks['run_client'].append(self.loop.create_task(self.internal_proxy_to_socket()))
+                self.tasks['run_client'].append(self.loop.create_task(self.socket_to_internal_proxy()))                
+                
+            await self.loop.sock_sendall(self.sock, proxy_msg)
+            resp = await self.loop.sock_recv(self.sock, 2048)
+            self.logger.info(f'Proxy server response received : {str(resp)}')
+            regex_status = b'HTTP\/\d.\d 200'
+            if re.match(regex_status, resp[:12]):
+            #if resp.startswith(b'HTTP/1.1 200') or resp.startswith(b'HTTP/1.0 200'):
+                self.logger.info('Proxy success')
+            else:
+                self.logger.error('Proxy failure, raising ConnectionRefusedError')
+                raise ConnectionRefusedError()
+        except Exception:
+            self.logger.exception('proxy_connect')
+            raise
+        
+    async def internal_proxy_to_socket(self):
+        #used only in ssl_server
+        #forward data received by self.outside_end_sockpair to self.sock
+        while True:
+            #data sent by client goes from inside_end_sockpair (in run_client) to outside_end_sockpair
+            #then we must forward it from self.outside_end_sockpair to self.sock :
+            try:
+                data = await self.loop.sock_recv(self.outside_end_sockpair, 2048)
+                await self.loop.sock_sendall(self.sock, data)
+            except Exception:
+                self.logger.exception('internal_proxy_to_socket')
+            
+    async def socket_to_internal_proxy(self):
+        #used only in ssl_server        
+        #forward data received by self.sock to self.outside_end_sockpair
+        while True:
+            #we must forward data received by client self.sock into outside_end_sockpair      
+            #then this data will reach self.inside_end_sockpair in run_client
+            try:
+                data = await self.loop.sock_recv(self.sock, 2048)
+                await self.loop.sock_sendall(self.outside_end_sockpair, data)
+            except Exception:
+                self.logger.exception('socket_to_internal_proxy')
+            
+        
     async def stop(self, connector_socket_only=False, client_wait_for_reconnect=False, hard=False,
                    shutdown=False, enable_delete_files=True):
             
@@ -489,7 +595,8 @@ class Connector:
     def cancel_tasks(self, task_names=None, task_excludes=None):
         #If task_names=None, cancel all tasks, else cancel task_names list of tasks
         try:
-            self.logger.info('cancel_tasks called. Current self.tasks are : '+str(list(self.tasks.keys())))
+            self.logger.info(f'cancel_tasks called with task_names : {task_names} and task_excludes : {task_excludes}'
+                             f' and current self.tasks : {list(self.tasks.keys())}')
             tasks_to_pop = []
             for name, tasks in self.tasks.items():
                 
@@ -551,8 +658,20 @@ class Connector:
                 if '.' not in server_sockaddr_addr:
                     #this is useful only when client and server are on the same machine                    
                     server_sockaddr_addr = iface_to_ip(server_sockaddr_addr)
-                await asyncio.wait_for(self.loop.sock_connect(self.sock,
-                                            (server_sockaddr_addr, self.server_sockaddr[1])), timeout=2)                              
+
+                if not self.proxy.get('enabled'):
+                    await asyncio.wait_for(self.loop.sock_connect(self.sock,
+                                                (server_sockaddr_addr, self.server_sockaddr[1])), timeout=2)                                
+                    self.logger.debug(f'Created socket for {self.source_id} with info {str(self.sock.getsockname())} '
+                                     f'to peer {self.sock.getpeername()}')
+                else:
+                    #only for client
+                    await asyncio.wait_for(self.loop.sock_connect(self.sock,
+                                                (self.proxy['address'], self.proxy['port'])), timeout=2)                                
+                    self.logger.debug(f'Created socket for {self.source_id} with info {str(self.sock.getsockname())} '
+                                     f'to proxy {self.sock.getpeername()}')
+                    await self.proxy_connect(self.server_sockaddr, server_sockaddr_addr=server_sockaddr_addr)
+                      
             except asyncio.CancelledError:
                 raise
             except Exception as exc:    #(ConnectionRefusedError, asyncio.TimeoutError):
@@ -597,7 +716,7 @@ class Connector:
                 if response:
                     response = str(response).encode()
                 else:
-                    response = b'success'    #b'{}'
+                    response = b'{}' #b'success'    
                 response = Structures.MSG_4_STRUCT.pack(len(response)) + response
                 self.logger.info(f'commander_cb responding : {response}')
                 writer.write(response)            
@@ -629,7 +748,7 @@ class Connector:
         #elif self.queue_send:
         #    res = list(sorted(self.queue_send.keys()))            
         if dump_result:
-            return str(res)
+            return json.dumps(res)
         return res
 
     def manage_ignore_peer_traffic(self, show=False, enable=False, disable=False, unique_peer=False):
@@ -668,13 +787,14 @@ class Connector:
             self.logger.exception('delete_client_certificate_on_server')
             return json.dumps({'status':False, 'msg':str(exc)})
 
-    async def delete_client_certificate_on_client(self):
+    async def delete_client_certificate_on_client(self, restart_client=True):
         try:
             self.logger.info(f'{self.source_id} deleting own certificate')            
             response = await self.ssl_helper.remove_client_cert_on_client(self.source_id)
             self.client_certificate_name = self.ssl_helper.CLIENT_DEFAULT_CERT_NAME
             self.logger.info('Client will use again the default certificate : '+self.client_certificate_name)
-            await self.restart()            
+            if restart_client:
+                await self.restart()            
             return response
         except Exception as exc:
             self.logger.exception('delete_client_certificate_on_client')
@@ -713,6 +833,38 @@ class Connector:
         except Exception:
             self.logger.exception(f'{self.source_id} delete_previous_persistence_remains')
                 
+    def show_subscribe_message_types(self):
+        self.logger.info(f'{self.source_id} show_subscribe_message_types')
+        if not self.is_server:
+            return self.subscribe_message_types
+        else:
+            return []
+
+    async def set_subscribe_message_types(self, message_types):
+        self.logger.info(f'{self.source_id} set_subscribe_message_types {message_types}')
+        message_types_to_remove = [message_type for message_type in message_types \
+                                   if message_type not in self.recv_message_types]
+        if message_types_to_remove:
+            self.logger.warning(f'Ignoring {message_types_to_remove} because they are not in recv_message_types')
+        message_types = list(set(message_types) - set(message_types_to_remove))
+        if not self.is_server:
+            try:
+                unsubscribe_to_message_types = set(self.subscribe_message_types) - set(message_types)
+                subscribe_to_message_types = set(message_types) - set(self.subscribe_message_types)
+                send_message = self.full_duplex_connections[str(self.server_sockaddr)].send_message
+                if unsubscribe_to_message_types:
+                    await send_message(message_type='_pubsub',
+                                data=json.dumps({'unsubscribe_message_types':list(unsubscribe_to_message_types)}))
+                if subscribe_to_message_types:
+                    await send_message(message_type='_pubsub',
+                                data=json.dumps({'subscribe_message_types':list(subscribe_to_message_types)}))                
+                self.subscribe_message_types = message_types
+            except Exception:
+                self.logger.exception('set_subscribe_message_types')
+            return self.subscribe_message_types
+        else:
+            return []
+            
     def store_message_to_persistence(self, peername, message, ignore_count=False):
         persistence_path = self.persistence_path + self.alnum_name(peername)
         try:
@@ -770,7 +922,7 @@ class Connector:
                 else:
                     while True:
                         #read file in chunks
-                        chunk = fd.read(self.READ_CHUNK_SIZE)
+                        chunk = fd.read(self.READ_PERSISTENCE_CHUNK_SIZE)
                         if not chunk:
                             last_iteration = True                    
                         chunk = last_element + chunk
@@ -818,7 +970,7 @@ class Connector:
     #4|2|json|4|data|4|binary
     def pack_message(self, transport_json=None, message_type=None, source_id=None, destination_id=None,
                      request_id=None, response_id=None, binary=None, await_response=False,
-                     with_file=None, data=None, wait_for_ack=False):
+                     with_file=None, data=None, wait_for_ack=False, message_type_publish=None):
         if DEBUG_SHOW_DATA:
             self.logger.debug('pack_message with params : '+str(message_type)+', '+str(data)+', '+str(transport_json))
         if transport_json is None:
@@ -839,6 +991,8 @@ class Connector:
                 transport_json[MessageFields.WITH_FILE] = with_file
             if wait_for_ack:
                 transport_json[MessageFields.WAIT_FOR_ACK] = wait_for_ack
+            if message_type_publish:
+                transport_json[MessageFields.MESSAGE_TYPE_PUBLISH] = message_type_publish
                 
         #pack message
         json_field = json.dumps(transport_json).encode()
@@ -973,7 +1127,8 @@ class Connector:
                         send_to_queue = False
                             
                 message_type = transport_json[MessageFields.MESSAGE_TYPE]
-                if (message_type not in self.send_message_types) and (message_type != '_ping'):
+                if (message_type not in self.send_message_types) and (message_type != '_ping') \
+                            and (message_type != '_pubsub') and (not self.pubsub_central_broker):
                     self.logger.warning(f'{self.source_id} queue_send_to_connector_put received a message with '
                                         f'invalid type {message_type}. Ignoring...')                    
                     send_to_queue = False                    
@@ -987,36 +1142,40 @@ class Connector:
                         request_id = uuid.uuid4().hex
                     if request_id in self.messages_awaiting_response[message_type].get(peername, {}):
                         self.logger.warning(f'Request id {request_id} for type {message_type} for peer {peername} '
-                                            'already in self.messages_awaiting_response, ignoring')
-                        send_to_queue = False
-                    else:
-                        if not send_to_queue:
-                            if self.disk_persistence or self.ram_persistence:
-                                self.logger.warning(f'{self.source_id} Connector is currently disconnected. '
-                                                    f'Response to request_id {request_id} will be sent when peer goes up'
-                                                    ', since persistence is enabled')
-                                await_response = True                        
-                            else:
-                                self.logger.warning(f'{self.source_id} Connector is currently disconnected. '
-                                                    f'Response to request_id {request_id} will not be sent since '
-                                                    'persistence is disabled')
+                                            'already in self.messages_awaiting_response, overriding it !')                       
+#                        self.logger.warning(f'Request id {request_id} for type {message_type} for peer {peername} '
+#                                            'already in self.messages_awaiting_response, ignoring')
+#                        send_to_queue = False
+#                    else:
+                    if not send_to_queue:
+                        if self.disk_persistence or self.ram_persistence:
+                            self.logger.warning(f'{self.source_id} Connector is currently disconnected. '
+                                                f'Response to request_id {request_id} will be sent when peer goes up'
+                                                ', since persistence is enabled')
+                            await_response = True                        
                         else:
-                            if len(self.messages_awaiting_response[message_type].get(peername, {})) > self.MAX_NUMBER_OF_AWAITING_REQUESTS:
-                                self.logger.error(f'{self.source_id} messages_awaiting_response dict has '
-                                                  f'{len(self.messages_awaiting_response[message_type].get(peername,{}))}'
-                                                  f' entries for message type {message_type} and peername {peername}. '
-                                                  'Stop adding entries until it goes '
-                                                  f'below {self.MAX_NUMBER_OF_AWAITING_REQUESTS}')
-                            else:
-                                self.logger.info(f'{self.source_id} Adding request {request_id} to '
-                                                 f'messages_awaiting_response dict for type {message_type} and '
-                                                 f'peer {peername}')
-                                await_response = True
-                        if await_response:
-                            if peername not in self.messages_awaiting_response[message_type]:
-                                self.messages_awaiting_response[message_type][peername] = {}
-                            #2nd element will contain the response set by handle_incoming_connection
-                            self.messages_awaiting_response[message_type][peername][request_id] = [asyncio.Event(),None]    
+                            self.logger.warning(f'{self.source_id} Connector is currently disconnected. '
+                                                f'Response to request_id {request_id} will not be sent since '
+                                                'persistence is disabled')
+                            #close writer so that client stops waiting uselessly
+                            writer.close()
+                    else:
+                        if len(self.messages_awaiting_response[message_type].get(peername, {})) > self.MAX_NUMBER_OF_AWAITING_REQUESTS:
+                            self.logger.error(f'{self.source_id} messages_awaiting_response dict has '
+                                              f'{len(self.messages_awaiting_response[message_type].get(peername,{}))}'
+                                              f' entries for message type {message_type} and peername {peername}. '
+                                              'Stop adding entries until it goes '
+                                              f'below {self.MAX_NUMBER_OF_AWAITING_REQUESTS}')
+                        else:
+                            self.logger.info(f'{self.source_id} Adding request {request_id} to '
+                                             f'messages_awaiting_response dict for type {message_type} and '
+                                             f'peer {peername}')
+                            await_response = True
+                    if await_response:
+                        if peername not in self.messages_awaiting_response[message_type]:
+                            self.messages_awaiting_response[message_type][peername] = {}
+                        #2nd element will contain the response set by handle_incoming_connection
+                        self.messages_awaiting_response[message_type][peername][request_id] = [asyncio.Event(),None]    
                             
                 if send_to_queue:            
                     queue_send = self.queue_send[peername]
@@ -1124,7 +1283,7 @@ class Connector:
                 else:
                     while True:
                         #read file in chunks
-                        chunk = fd.read(self.READ_CHUNK_SIZE)
+                        chunk = fd.read(self.READ_PERSISTENCE_CHUNK_SIZE)
                         if not chunk:
                             last_iteration = True                    
                         chunk = last_element + chunk
@@ -1394,7 +1553,9 @@ class Connector:
         try:
             ssl_context = self.build_client_ssl_context() if self.use_ssl else None        
             server_hostname = '' if self.use_ssl else None
-            reader, writer = await asyncio.wait_for(asyncio.open_connection(sock=self.sock, ssl=ssl_context,
+            #self.inside_end_sockpair is not None only when proxy with ssl_server is configured
+            socket_used = self.inside_end_sockpair or self.sock
+            reader, writer = await asyncio.wait_for(asyncio.open_connection(sock=socket_used, ssl=ssl_context,
                                     server_hostname=server_hostname, limit=self.MAX_SOCKET_BUFFER_SIZE), 
                                     timeout=self.ASYNC_TIMEOUT)                                                                     
             await self.manage_full_duplex(reader, writer)
