@@ -5,9 +5,16 @@ import subprocess
 import uuid
 import shutil
 
+from .helpers import get_tmp_dir
 
+def update_conf(conf_template, conf, replacement_dict):
+    shutil.copy(conf_template, conf)
+    with open(conf, 'a') as fd:
+        for key,value in replacement_dict.items():
+            fd.write(str(key)+' = '+str(value)+'\n')
+        
 class SSL_helper:
-    DEFAULT_BASE_PATH = '/tmp/aioconnectors' #os.getcwd()
+    DEFAULT_BASE_PATH = get_tmp_dir() #os.getcwd()
     CLIENT_DEFAULT_CERT_NAME = 'default'    
     SOURCE_ID_2_CERT = 'source_id_2_cert.json'
     CERT_NAME_EXTENSION = "pem"
@@ -32,6 +39,7 @@ class SSL_helper:
             self.SERVER_PEM_PATH = os.path.join(self.SERVER_BASE_PATH, 'server-cert/server.'+self.CERT_NAME_EXTENSION)
             self.SERVER_KEY_PATH = os.path.join(self.SERVER_BASE_PATH, 'server-cert/server.'+self.KEY_NAME_EXTENSION)
             self.CSR_CONF  = os.path.join(self.SERVER_BASE_PATH, 'csr_details.conf')
+            self.CSR_TEMPLATE_CONF  = os.path.join(self.SERVER_BASE_PATH, 'csr_details_template.conf')
             
             #client
             self.CLIENT_BASE_PATH = os.path.join(self.certificates_base_path, 'client')
@@ -89,13 +97,15 @@ class SSL_helper:
                 raise Exception(f'A certificate already exists for client {source_id}. '
                                 f'Use delete_client_certificate to delete it')
             if common_name:
-                create_certificate_cmd = f"openssl req -new -newkey rsa -nodes -x509 -days 3650 -subj '/O={common_name}/CN={common_name}' -keyout "\
+                update_conf(self.CSR_TEMPLATE_CONF, self.CSR_CONF, {'O':common_name, 'CN':common_name})
+                create_certificate_cmd = f"openssl req -new -newkey rsa -nodes -x509 -days 3650 -keyout "\
                                          f"{key_path} -out {crt_path} -config {self.CSR_CONF}"
             else:
                 #necessary to set a unique field (like organization), so that each certificate has a unique hash, 
                 #which is better for fast authentication
                 organization = uuid.uuid4().hex
-                create_certificate_cmd = f"openssl req -new -newkey rsa -nodes -x509 -days 3650 -subj '/O={organization}' -keyout "\
+                update_conf(self.CSR_TEMPLATE_CONF, self.CSR_CONF, {'O':organization})                
+                create_certificate_cmd = f"openssl req -new -newkey rsa -nodes -x509 -days 3650 -keyout "\
                                          f"{key_path} -out {crt_path} -config {self.CSR_CONF}"
                                      
             proc, stdout, stderr = await self.run_cmd(create_certificate_cmd)
@@ -247,9 +257,6 @@ def create_certificates(logger, certificates_directory_path):
     #certificates_path = os.path.join(ssl_helper.BASE_PATH, 'certificates')
     certificates_path = ssl_helper.certificates_base_path
     logger.info('Certificates will be created under directory : '+certificates_path)
-    if os.path.exists(certificates_path) and os.listdir(certificates_path):
-        logger.error(certificates_path+' should be empty before starting this process')
-        return False
         
     certificates_path_server = os.path.join(certificates_path, 'server')
     certificates_path_server_client = os.path.join(certificates_path_server, 'client-certs')       
@@ -258,6 +265,16 @@ def create_certificates(logger, certificates_directory_path):
     certificates_path_client = os.path.join(certificates_path, 'client')        
     certificates_path_client_client = os.path.join(certificates_path_client, 'client-certs')        
     certificates_path_client_server = os.path.join(certificates_path_client, 'server-cert')                
+
+    if os.path.exists(certificates_path_server_server) and os.listdir(certificates_path_server_server):
+        logger.error(certificates_path_server_server+' should be empty before starting this process')
+        return False
+    if os.path.exists(certificates_path_server_client) and os.listdir(certificates_path_server_client):
+        logger.error(certificates_path_server_client+' should be empty before starting this process')
+        return False
+    if os.path.exists(certificates_path_client) and os.listdir(certificates_path_client):
+        logger.error(certificates_path_client+' should be empty before starting this process')
+        return False
     
     if not os.path.exists(certificates_path_server_server):
         os.makedirs(certificates_path_server_server)
@@ -277,22 +294,27 @@ def create_certificates(logger, certificates_directory_path):
     SERVER_CERTS_PATH = os.path.join(certificates_path_server_client, '{}.pem')
     
     csr_details_conf = os.path.join(certificates_path_server, 'csr_details.conf')
+    csr_details_template_conf = os.path.join(certificates_path_server, 'csr_details_template.conf')
     
-    with open(csr_details_conf, 'w') as fd:
-        fd.write(
-            '''
-            [req]
-            prompt = no
-            default_bits = 2048
-            default_md = sha256
-            distinguished_name = dn
-            
-            [ dn ]
-            C=US
-            O=Company)
-            ''')
-    
+    #this if lets the user tweak the base csr_details_template_conf
+    if not os.path.exists(csr_details_template_conf):
+        with open(csr_details_template_conf, 'w') as fd:
+            fd.write(
+'''
+[req]
+prompt = no
+default_bits = 2048
+default_md = sha256
+distinguished_name = dn
+
+[ dn ]
+C = US
+''')
+    else:
+        logger.info(f'Using preexisting {csr_details_template_conf}')
+        
 #        1) Create Server certificate            
+    update_conf(csr_details_template_conf, csr_details_conf, {'O':'company'})                        
     cmd = f'openssl req -new -newkey rsa -nodes -x509 -days 3650 -keyout {SERVER_KEY_PATH} -out {SERVER_PEM_PATH} -config {csr_details_conf}'
     stdout = subprocess.check_output(cmd, shell=True)
     shutil.copy(SERVER_PEM_PATH, CLIENT_SERVER_CRT_PATH)
@@ -304,7 +326,8 @@ def create_certificates(logger, certificates_directory_path):
     organization = ssl_helper.CLIENT_DEFAULT_CERT_NAME
     #we might want to obfuscate organization
     organization = str(abs(hash(organization)) % (10 ** 8))
-    cmd = f"openssl req -new -newkey rsa -nodes -x509 -days 3650 -subj '/O={organization}' -keyout {client_default_key} -out {client_default_pem} -config {csr_details_conf}"
+    update_conf(csr_details_template_conf, csr_details_conf, {'O':organization})                    
+    cmd = f"openssl req -new -newkey rsa -nodes -x509 -days 3650 -keyout {client_default_key} -out {client_default_pem} -config {csr_details_conf}"
     stdout = subprocess.check_output(cmd, shell=True)        
     shutil.copy(client_default_pem, SERVER_CERTS_PATH.format(ssl_helper.CLIENT_DEFAULT_CERT_NAME))
 
@@ -322,4 +345,4 @@ def create_certificates(logger, certificates_directory_path):
     stdout = subprocess.check_output(cmd, shell=True)
     logger.info('Finished create_certificates')
     return True
-            
\ No newline at end of file
+            
